@@ -2,6 +2,7 @@
 """
 from __future__ import unicode_literals
 from abc import ABCMeta, abstractmethod
+from six import with_metaclass
 
 __all__ = (
     'Completion',
@@ -11,7 +12,8 @@ __all__ = (
 
 
 class Completion(object):
-    def __init__(self, text, start_position=0, display=None, display_meta=''):
+    def __init__(self, text, start_position=0, display=None, display_meta=None,
+                 get_display_meta=None):
         """
         :param text: The new string that will be inserted into the document.
         :param start_position: Position relative to the cursor_position where the
@@ -21,10 +23,13 @@ class Completion(object):
             differently in the completion menu.
         :param display_meta: (Optional string) Meta information about the
             completion, e.g. the path or source where it's coming from.
+        :param get_display_meta: Lazy `display_meta`. Retrieve meta information
+            only when meta is displayed.
         """
         self.text = text
         self.start_position = start_position
-        self.display_meta = display_meta
+        self._display_meta = display_meta
+        self._get_display_meta = get_display_meta
 
         if display is None:
             self.display = text
@@ -34,7 +39,8 @@ class Completion(object):
         assert self.start_position <= 0
 
     def __repr__(self):
-        return 'Completion(text=%r, start_position=%r)' % (self.text, self.start_position)
+        return '%s(text=%r, start_position=%r)' % (
+            self.__class__.__name__, self.text, self.start_position)
 
     def __eq__(self, other):
         return (
@@ -46,6 +52,19 @@ class Completion(object):
     def __hash__(self):
         return hash((self.text, self.start_position, self.display, self.display_meta))
 
+    @property
+    def display_meta(self):
+        # Return meta-text. (This is lazy when using "get_display_meta".)
+        if self._display_meta is not None:
+            return self._display_meta
+
+        elif self._get_display_meta:
+            self._display_meta = self._get_display_meta()
+            return self._display_meta
+
+        else:
+            return ''
+
 
 class CompleteEvent(object):
     """
@@ -60,16 +79,18 @@ class CompleteEvent(object):
         #: Used explicitely requested completion by pressing 'tab'.
         self.completion_requested = completion_requested
 
+    def __repr__(self):
+        return '%s(text_inserted=%r, completion_requested=%r)' % (
+            self.__class__.__name__, self.text_inserted, self.completion_requested)
 
-class Completer(object):
+
+class Completer(with_metaclass(ABCMeta, object)):
     """
     Base class for Code implementations.
 
     The methods in here are methods that are expected to exist for the `Buffer`
     and `Renderer` classes.
     """
-    __metaclass__ = ABCMeta
-
     @abstractmethod
     def get_completions(self, document, complete_event):
         """
@@ -79,13 +100,10 @@ class Completer(object):
             yield
 
 
-def get_common_complete_suffix(completer, document, complete_event):
+def get_common_complete_suffix(document, completions):
     """
-    return one `Completion` instance or None.
+    Return the common prefix for all completions.
     """
-    # If there is one completion, return that.
-    completions = list(completer.get_completions(document, complete_event))
-
     # Take only completions that don't change the text before the cursor.
     def doesnt_change_before_cursor(completion):
         end = completion.text[:-completion.start_position]
