@@ -3,6 +3,8 @@
 from __future__ import unicode_literals
 
 import re
+import six
+import string
 
 from .selection import SelectionType
 
@@ -13,15 +15,15 @@ __all__ = ('Document',)
 # characters a word, but also a group of special characters a word, as long as
 # it doesn't contain a space.)
 # (This is a 'word' in Vi.)
-_FIND_WORD_RE = re.compile('([a-zA-Z0-9_]+|[^a-zA-Z0-9_\s]+)')
-_FIND_CURRENT_WORD_RE = re.compile('^([a-zA-Z0-9_]+|[^a-zA-Z0-9_\s]+)')
-_FIND_CURRENT_WORD_INCLUDE_TRAILING_WHITESPACE_RE = re.compile('^(([a-zA-Z0-9_]+|[^a-zA-Z0-9_\s]+)\s*)')
+_FIND_WORD_RE = re.compile(r'([a-zA-Z0-9_]+|[^a-zA-Z0-9_\s]+)')
+_FIND_CURRENT_WORD_RE = re.compile(r'^([a-zA-Z0-9_]+|[^a-zA-Z0-9_\s]+)')
+_FIND_CURRENT_WORD_INCLUDE_TRAILING_WHITESPACE_RE = re.compile(r'^(([a-zA-Z0-9_]+|[^a-zA-Z0-9_\s]+)\s*)')
 
 # Regex for finding "WORDS" in documents.
 # (This is a 'WORD in Vi.)
-_FIND_BIG_WORD_RE = re.compile('([^\s]+)')
-_FIND_CURRENT_BIG_WORD_RE = re.compile('^([^\s]+)')
-_FIND_CURRENT_BIG_WORD_INCLUDE_TRAILING_WHITESPACE_RE = re.compile('^([^\s]+\s*)')
+_FIND_BIG_WORD_RE = re.compile(r'([^\s]+)')
+_FIND_CURRENT_BIG_WORD_RE = re.compile(r'^([^\s]+)')
+_FIND_CURRENT_BIG_WORD_INCLUDE_TRAILING_WHITESPACE_RE = re.compile(r'^([^\s]+\s*)')
 
 
 class Document(object):
@@ -39,6 +41,13 @@ class Document(object):
     __slots__ = ('text', 'cursor_position', 'selection')
 
     def __init__(self, text='', cursor_position=None, selection=None):
+        assert isinstance(text, six.text_type), 'Got %r' % text
+
+        # Check cursor position. It can also be right after the end. (Where we
+        # insert text.)
+        assert cursor_position is None or cursor_position <= len(text), AssertionError(
+                'cursor_position=%r, len_text=%r' % (cursor_position, len(text)))
+
         # By default, if no cursor position was given, make sure to put the
         # cursor position is at the end of the document. This is what makes
         # sense in most places.
@@ -49,6 +58,9 @@ class Document(object):
         self.cursor_position = cursor_position
         self.selection = selection
 
+    def __repr__(self):
+        return '%s(%r, %r)' % (self.__class__.__name__, self.text, self.cursor_position)
+
     @property
     def current_char(self):
         """ Return character under cursor, or None """
@@ -119,6 +131,20 @@ class Document(object):
         except IndexError:
             return ''
 
+    @property
+    def on_first_line(self):
+        """
+        True when we are at the first line.
+        """
+        return self.cursor_position_row == 0
+
+    @property
+    def on_last_line(self):
+        """
+        True when we are at the last line.
+        """
+        return self.cursor_position_row == self.line_count - 1
+
     @property
     def cursor_position_row(self):
         """
@@ -149,7 +175,12 @@ class Document(object):
         Given a (row, col) tuple, return the corresponding index.
         (Row and col params are 0-based.)
         """
-        return len('\n'.join(self.lines[:row])) + len('\n') + col
+        result = len('\n'.join(self.lines[:row])) + (len('\n') if row > 0 else 0) + col
+
+        # Keep in range. (len(self.text) is included, because the cursor can be
+        # right after the end of the text as well.)
+        result = max(0, min(result, len(self.text)))
+        return result
 
     @property
     def is_cursor_at_the_end(self):
@@ -167,13 +198,16 @@ class Document(object):
         """
         return self.text[self.cursor_position:].find(sub) == 0
 
-    def find(self, sub, in_current_line=False, include_current_position=False, count=1):  # TODO: rename to `find_forwards`
+    def find(self, sub, in_current_line=False, include_current_position=False,
+            ignore_case=False, count=1):
         """
         Find `text` after the cursor, return position relative to the cursor
         position. Return `None` if nothing was found.
 
         :param count: Find the n-th occurance.
         """
+        assert isinstance(ignore_case, bool)
+
         if in_current_line:
             text = self.current_line_after_cursor
         else:
@@ -185,7 +219,8 @@ class Document(object):
             else:
                 text = text[1:]
 
-        iterator = re.finditer(re.escape(sub), text)
+        flags = re.IGNORECASE if ignore_case else 0
+        iterator = re.finditer(re.escape(sub), text, flags)
 
         try:
             for i, match in enumerate(iterator):
@@ -197,14 +232,15 @@ class Document(object):
         except StopIteration:
             pass
 
-    def find_all(self, sub):
+    def find_all(self, sub, ignore_case=False):
         """
         Find all occurances of the substring. Return a list of absolute
         positions in the document.
         """
-        return [a.start() for a in re.finditer(re.escape(sub), self.text)]
+        flags = re.IGNORECASE if ignore_case else 0
+        return [a.start() for a in re.finditer(re.escape(sub), self.text, flags)]
 
-    def find_backwards(self, sub, in_current_line=False, count=1):
+    def find_backwards(self, sub, in_current_line=False, ignore_case=False, count=1):
         """
         Find `text` before the cursor, return position relative to the cursor
         position. Return `None` if nothing was found.
@@ -216,7 +252,8 @@ class Document(object):
         else:
             before_cursor = self.text_before_cursor[::-1]
 
-        iterator = re.finditer(re.escape(sub[::-1]), before_cursor)
+        flags = re.IGNORECASE if ignore_case else 0
+        iterator = re.finditer(re.escape(sub[::-1]), before_cursor, flags)
 
         try:
             for i, match in enumerate(iterator):
@@ -254,7 +291,8 @@ class Document(object):
         except StopIteration:
             pass
 
-    def find_boundaries_of_current_word(self, WORD=False, include_leading_whitespace=False, include_trailing_whitespace=False):
+    def find_boundaries_of_current_word(self, WORD=False, include_leading_whitespace=False,
+                                        include_trailing_whitespace=False):
         """
         Return the relative boundaries (startpos, endpos) of the current word under the
         cursor. (This is at the current line, because line boundaries obviously
@@ -275,11 +313,31 @@ class Document(object):
         match_before = get_regex(include_leading_whitespace).search(text_before_cursor)
         match_after = get_regex(include_trailing_whitespace).search(text_after_cursor)
 
+        # When there is a match before and after, and we're not looking for
+        # WORDs, make sure that both the part before and after the cursor are
+        # either in the [a-zA-Z_] alphabet or not. Otherwise, drop the part
+        # before the cursor.
+        if not WORD and match_before and match_after:
+            c1 = self.text[self.cursor_position - 1]
+            c2 = self.text[self.cursor_position]
+            alphabet = string.ascii_letters + '0123456789_'
+
+            if (c1 in alphabet) != (c2 in alphabet):
+                match_before = None
+
         return (
             - match_before.end(1) if match_before else 0,
             match_after.end(1) if match_after else 0
         )
 
+    def get_word_under_cursor(self, WORD=False):
+        """
+        Return the word, currently below the cursor.
+        This returns an empty string when the cursor is on a whitespace region.
+        """
+        start, end = self.find_boundaries_of_current_word(WORD=WORD)
+        return self.text[self.cursor_position + start: self.cursor_position + end]
+
     def find_next_word_beginning(self, count=1, WORD=False):
         """
         Return an index relative to the cursor position pointing to the start
@@ -340,23 +398,39 @@ class Document(object):
         except StopIteration:
             pass
 
-    def find_next_matching_line(self, match_func):
+    def find_next_matching_line(self, match_func, count=1):
         """
         Look downwards for empty lines.
         Return the line index, relative to the current line.
         """
+        result = None
+
         for index, line in enumerate(self.lines[self.cursor_position_row + 1:]):
             if match_func(line):
-                return 1 + index
+                result = 1 + index
+                count -= 1
 
-    def find_previous_matching_line(self, match_func):
+            if count == 0:
+                break
+
+        return result
+
+    def find_previous_matching_line(self, match_func, count=1):
         """
         Look upwards for empty lines.
         Return the line index, relative to the current line.
         """
+        result = None
+
         for index, line in enumerate(self.lines[:self.cursor_position_row][::-1]):
             if match_func(line):
-                return -1 - index
+                result = -1 - index
+                count -= 1
+
+            if count == 0:
+                break
+
+        return result
 
     def get_cursor_left_position(self, count=1):
         """
@@ -370,7 +444,7 @@ class Document(object):
         """
         return min(count, len(self.current_line_after_cursor))
 
-    def get_cursor_up_position(self, count=1):  # TODO: implement `count`
+    def get_cursor_up_position(self, count=1):
         """
         Return the relative cursor position (character index) where we would be if the
         user pressed the arrow-up button.
@@ -453,18 +527,16 @@ class Document(object):
 
         return 0
 
-    @property
-    def home_position(self):
+    def get_start_of_document_position(self):
         """ Relative position for the start of the document. """
         return - self.cursor_position
 
-    @property
-    def end_position(self):
+    def get_end_of_document_position(self):
         """ Relative position for the end of the document. """
         return len(self.text) - self.cursor_position
 
     def get_start_of_line_position(self, after_whitespace=False):
-        """ Relative position for the end of this line. """
+        """ Relative position for the start of this line. """
         if after_whitespace:
             current_line = self.current_line
             return len(current_line) - len(current_line.lstrip()) - self.cursor_position_col
