@@ -1,14 +1,17 @@
 from __future__ import unicode_literals
 from functools import wraps
+from prompt_toolkit.filters import CLIFilter, Always
 
 
-def create_handle_decorator(registry, filter):
+def create_handle_decorator(registry, filter=Always()):
     """
     Create a key handle decorator, which is compatible with `Registry.handle`
     but has a `save_before` option, which will make sure that undo changes are
     saved to the undo stack of the `Buffer` object before every key press
     event.
     """
+    assert isinstance(filter, CLIFilter)
+
     def handle(*keys, **kw):
         safe_before = kw.pop('save_before', True)
 
