@@ -1,12 +1,14 @@
+# pylint: disable=function-redefined
 from __future__ import unicode_literals
-from prompt_toolkit.enums import IncrementalSearchDirection
-from prompt_toolkit.keys import Keys
 from prompt_toolkit.buffer import ClipboardData, indent, unindent
-from prompt_toolkit.selection import SelectionType
+from prompt_toolkit.document import Document
+from prompt_toolkit.enums import IncrementalSearchDirection, SEARCH_BUFFER, SYSTEM_BUFFER
+from prompt_toolkit.filters import Filter, Condition, HasArg, Always, to_cli_filter
 from prompt_toolkit.key_binding.vi_state import ViState, CharacterFind, InputMode
-from prompt_toolkit.filters import Filter
+from prompt_toolkit.keys import Keys
+from prompt_toolkit.layout.utils import find_window_for_buffer_name
+from prompt_toolkit.selection import SelectionType
 
-from .basic import load_basic_bindings
 from .utils import create_handle_decorator
 
 import prompt_toolkit.filters as filters
@@ -34,6 +36,7 @@ class ViStateFilter(Filter):
 class CursorRegion(object):
     """
     Return struct for functions wrapped in ``change_delete_move_yank_handler``.
+    Bot `start` and `end` are relative to the current cursor position.
     """
     def __init__(self, start, end=0):
         self.start = start
@@ -49,16 +52,19 @@ class CursorRegion(object):
             return self.end, self.start
 
 
-def load_vi_bindings(registry, vi_state, filter=None):
+def load_vi_bindings(registry, vi_state, enable_visual_key=Always(), filter=None):
     """
     Vi extensions.
 
     # Overview of Readline Vi commands:
     # http://www.catonmat.net/download/bash-vi-editing-mode-cheat-sheet.pdf
+
+    :param enable_visual_key: Filter to enable lowercase 'v' bindings. A reason to disable these
+         are to support open-in-editor functionality. These key bindings conflict.
     """
     assert isinstance(vi_state, ViState)
+    enable_visual_key = to_cli_filter(enable_visual_key)
 
-    load_basic_bindings(registry, filter)
     handle = create_handle_decorator(registry, filter)
 
     insert_mode = ViStateFilter(vi_state, InputMode.INSERT) & ~ filters.HasSelection()
@@ -97,7 +103,7 @@ def load_vi_bindings(registry, vi_state, filter=None):
                 len(buffer.document.current_line) > 0):
             buffer.cursor_position -= 1
 
-    registry.onHandlerCalled += check_cursor_position
+    registry.on_handler_called += check_cursor_position
 
     @handle(Keys.Escape)
     def _(event):
@@ -114,20 +120,6 @@ def load_vi_bindings(registry, vi_state, filter=None):
         if bool(buffer.selection_state):
             buffer.exit_selection()
 
-    @handle('k', filter=navigation_mode)
-    def _(event):
-        """
-        Arrow up in navigation mode.
-        """
-        event.current_buffer.auto_up(count=event.arg)
-
-    @handle('j', filter=navigation_mode)
-    def _(event):
-        """
-        Arrow down in navigation mode.
-        """
-        event.current_buffer.auto_down(count=event.arg)
-
     @handle('k', filter=selection_mode)
     def _(event):
         """
@@ -142,19 +134,25 @@ def load_vi_bindings(registry, vi_state, filter=None):
         """
         event.current_buffer.cursor_down(count=event.arg)
 
+    @handle('k', filter=navigation_mode)
     @handle(Keys.Up, filter=navigation_mode)
+    @handle(Keys.ControlP, filter=navigation_mode)
     def _(event):
         """
-        Arrow up in navigation mode.
+        Arrow up and ControlP in navigation mode go up.
         """
-        event.current_buffer.auto_up(count=event.arg)
+        b = event.current_buffer
+        b.auto_up(count=event.arg)
 
+    @handle('j', filter=navigation_mode)
     @handle(Keys.Down, filter=navigation_mode)
+    @handle(Keys.ControlN, filter=navigation_mode)
     def _(event):
         """
-        Arrow down in navigation mode.
+        Arrow down and Control-N in navigation mode.
         """
-        event.current_buffer.auto_down(count=event.arg)
+        b = event.current_buffer
+        b.auto_down(count=event.arg)
 
     @handle(Keys.Backspace, filter=navigation_mode)
     def _(event):
@@ -173,21 +171,24 @@ def load_vi_bindings(registry, vi_state, filter=None):
 
     @handle(Keys.ControlN, filter=insert_mode)
     def _(event):
-        event.current_buffer.complete_next()
+        b = event.current_buffer
 
-    @handle(Keys.ControlN, filter=navigation_mode)
-    def _(event):
-        """
-        Control-N: Next completion.
-        """
-        event.current_buffer.auto_down()
+        if b.complete_state:
+            b.complete_next()
+        else:
+            event.cli.start_completion(select_first=True)
 
     @handle(Keys.ControlP, filter=insert_mode)
     def _(event):
         """
         Control-P: To previous completion.
         """
-        event.current_buffer.complete_previous()
+        b = event.current_buffer
+
+        if b.complete_state:
+            b.complete_previous()
+        else:
+            event.cli.start_completion(select_last=True)
 
     @handle(Keys.ControlY, filter=insert_mode)
     def _(event):
@@ -203,21 +204,15 @@ def load_vi_bindings(registry, vi_state, filter=None):
         """
         event.current_buffer.cancel_completion()
 
-    @handle(Keys.ControlP, filter=navigation_mode)
-    def _(event):
-        """
-        CtrlP in navigation mode goes up.
-        """
-        event.current_buffer.auto_up()
-
     @handle(Keys.ControlJ, filter=navigation_mode)
     def _(event):
         """
         In navigation mode, pressing enter will always return the input.
         """
-        if event.current_buffer.validate():
-            event.current_buffer.add_to_history()
-            event.cli.set_return_value(event.current_buffer.document)
+        b = event.current_buffer
+
+        if b.accept_action.is_returnable:
+            b.accept_action.validate_and_handle(event.cli, b)
 
     # ** In navigation mode **
 
@@ -286,30 +281,15 @@ def load_vi_bindings(registry, vi_state, filter=None):
         if before and after:
             before = before + '\n'
 
-        buffer.text = before + after
-
-        # Set cursor position. (At the start of the first 'after' line, after the leading whitespace.)
-        buffer.cursor_position = len(before) + len(after) - len(after.lstrip(' '))
+        # Set text and cursor position.
+        buffer.document = Document(
+            text=before + after,
+            # Cursor At the start of the first 'after' line, after the leading whitespace.
+            cursor_position = len(before) + len(after) - len(after.lstrip(' ')))
 
         # Set clipboard data
         event.cli.clipboard.set_data(ClipboardData(deleted, SelectionType.LINES))
 
-    @handle('G', filter=navigation_mode)
-    def _(event):
-        """
-        If an argument is given, move to this line in the  history. (for
-        example, 15G) Otherwise, go the the last line of the current string.
-        """
-        buffer = event.current_buffer
-
-        # If an arg has been given explicitely.
-        if event._arg:
-            buffer.go_to_history(event.arg - 1)
-
-        # Otherwise this goes to the last line of the file.
-        else:
-            buffer.cursor_position = len(buffer.text)
-
     @handle('i', filter=navigation_mode)
     def _(event):
         vi_state.input_mode = InputMode.INSERT
@@ -321,25 +301,32 @@ def load_vi_bindings(registry, vi_state, filter=None):
 
     @handle('J', filter=navigation_mode)
     def _(event):
+        """ Join lines. """
         for i in range(event.arg):
             event.current_buffer.join_next_line()
 
+    @handle('J', filter=selection_mode)
+    def _(event):
+        """ Join selected lines. """
+        event.current_buffer.join_selected_lines()
+
     @handle('n', filter=navigation_mode)
-    def _(event):  # XXX: use `change_delete_move_yank_handler` and implement 'arg'
+    def _(event):  # XXX: use `change_delete_move_yank_handler`
         """
         Search next.
         """
-        event.current_buffer.incremental_search(vi_state.search_direction)
+        event.current_buffer.apply_search(
+            event.cli.search_state, include_current_position=False,
+            count=event.arg)
 
     @handle('N', filter=navigation_mode)
-    def _(event):  # TODO: use `change_delete_move_yank_handler` and implement 'arg'
+    def _(event):  # TODO: use `change_delete_move_yank_handler`
         """
         Search previous.
         """
-        if vi_state.search_direction == IncrementalSearchDirection.BACKWARD:
-            event.current_buffer.incremental_search(IncrementalSearchDirection.FORWARD)
-        else:
-            event.current_buffer.incremental_search(IncrementalSearchDirection.BACKWARD)
+        event.current_buffer.apply_search(
+            ~event.cli.search_state, include_current_position=False,
+            count=event.arg)
 
     @handle('p', filter=navigation_mode)
     def _(event):
@@ -385,22 +372,11 @@ def load_vi_bindings(registry, vi_state, filter=None):
         event.cli.clipboard.set_text(text)
         vi_state.input_mode = InputMode.INSERT
 
-    @handle('u', filter=navigation_mode)
+    @handle('u', filter=navigation_mode, save_before=False)
     def _(event):
         for i in range(event.arg):
             event.current_buffer.undo()
 
-    @handle('v', filter=navigation_mode)
-    def _(event):
-        event.current_buffer.open_in_editor()
-
-    # @handle('v', filter=navigation_mode)
-    # def _(event):
-    #     """
-    #     Start characters selection.
-    #     """
-    #     event.current_buffer.start_selection(selection_type=SelectionType.CHARACTERS)
-
     @handle('V', filter=navigation_mode)
     def _(event):
         """
@@ -428,7 +404,7 @@ def load_vi_bindings(registry, vi_state, filter=None):
         event.cli.clipboard.set_text(text)
 
     @handle('x', filter=selection_mode)
-    @handle('d', 'd', filter=selection_mode)
+    @handle('d', filter=selection_mode)
     def _(event):
         """
         Cut selection.
@@ -441,11 +417,8 @@ def load_vi_bindings(registry, vi_state, filter=None):
         """
         Change selection (cut and go to insert mode).
         """
-        buffer = event.current_buffer
-
-        selection_type = buffer.selection_state.type
-        deleted = buffer.cut_selection()
-        event.cli.clipboard.set_data(ClipboardData(deleted, selection_type))
+        clipboard_data = event.current_buffer.cut_selection()
+        event.cli.clipboard.set_data(clipboard_data)
         vi_state.input_mode = InputMode.INSERT
 
     @handle('y', filter=selection_mode)
@@ -540,7 +513,8 @@ def load_vi_bindings(registry, vi_state, filter=None):
         """
         Open line above and enter insertion mode
         """
-        event.current_buffer.insert_line_above()
+        event.current_buffer.insert_line_above(
+                copy_margin=not event.cli.in_paste_mode)
         vi_state.input_mode = InputMode.INSERT
 
     @handle('o', filter=navigation_mode)
@@ -548,7 +522,8 @@ def load_vi_bindings(registry, vi_state, filter=None):
         """
         Open line below and enter insertion mode
         """
-        event.current_buffer.insert_line_below()
+        event.current_buffer.insert_line_below(
+                copy_margin=not event.cli.in_paste_mode)
         vi_state.input_mode = InputMode.INSERT
 
     @handle('~', filter=navigation_mode)
@@ -568,14 +543,28 @@ def load_vi_bindings(registry, vi_state, filter=None):
         """
         Go to previous occurence of this word.
         """
-        pass
+        b = event.cli.current_buffer
+
+        search_state = event.cli.search_state
+        search_state.text = b.document.get_word_under_cursor()
+        search_state.direction = IncrementalSearchDirection.BACKWARD
+
+        b.apply_search(search_state, count=event.arg,
+                       include_current_position=False)
 
     @handle('*', filter=navigation_mode)
     def _(event):
         """
         Go to next occurence of this word.
         """
-        pass
+        b = event.cli.current_buffer
+
+        search_state = event.cli.search_state
+        search_state.text = b.document.get_word_under_cursor()
+        search_state.direction = IncrementalSearchDirection.FORWARD
+
+        b.apply_search(search_state, count=event.arg,
+                       include_current_position=False)
 
     @handle('(', filter=navigation_mode)
     def _(event):
@@ -593,6 +582,9 @@ def load_vi_bindings(registry, vi_state, filter=None):
         The decorated function should return a ``CursorRegion``.
         This decorator will create both the 'change', 'delete' and move variants,
         based on that ``CursorRegion``.
+
+        When there is nothing selected yet, this will also handle the "visual"
+        binding. E.g. 'viw' should select the current word.
         """
         no_move_handler = kw.pop('no_move_handler', False)
 
@@ -601,8 +593,7 @@ def load_vi_bindings(registry, vi_state, filter=None):
         #  See: :help motion.txt
         def decorator(func):
             if not no_move_handler:
-                @handle(*keys, filter=navigation_mode)
-                @handle(*keys, filter=selection_mode)
+                @handle(*keys, filter=navigation_mode|selection_mode)
                 def move(event):
                     """ Create move handler. """
                     region = func(event)
@@ -640,6 +631,19 @@ def load_vi_bindings(registry, vi_state, filter=None):
                 if substring:
                     event.cli.clipboard.set_text(substring)
 
+            @handle('v', *keys, filter=navigation_mode & enable_visual_key)
+            def visual_handler(event):
+                """ Create visual handler. (Enter character selection mode.) """
+                region = func(event)
+                buffer = event.current_buffer
+
+                start, end = region.sorted()
+                end += buffer.cursor_position - 1
+
+                buffer.cursor_position += start
+                buffer.start_selection(selection_type=SelectionType.CHARACTERS)
+                buffer.cursor_position = end
+
             def create(delete_only):
                 """ Create delete and change handlers. """
                 @handle('cd'[delete_only], *keys, filter=navigation_mode)
@@ -690,13 +694,13 @@ def load_vi_bindings(registry, vi_state, filter=None):
     def _(event):
         """ 'word' forward. 'cw', 'dw', 'w': Delete/change/move one word.  """
         return CursorRegion(event.current_buffer.document.find_next_word_beginning(count=event.arg) or
-                            event.current_buffer.document.end_position)
+                            event.current_buffer.document.get_end_of_document_position())
 
     @change_delete_move_yank_handler('W')
     def _(event):
         """ 'WORD' forward. 'cW', 'dW', 'W': Delete/change/move one WORD.  """
         return CursorRegion(event.current_buffer.document.find_next_word_beginning(count=event.arg, WORD=True) or
-                            event.current_buffer.document.end_position)
+                            event.current_buffer.document.get_end_of_document_position())
 
     @change_delete_move_yank_handler('e')
     def _(event):
@@ -757,26 +761,32 @@ def load_vi_bindings(registry, vi_state, filter=None):
         @change_delete_move_yank_handler('ai'[inner], ci_start, no_move_handler=True)
         @change_delete_move_yank_handler('ai'[inner], ci_end, no_move_handler=True)
         def _(event):
-            start = event.current_buffer.document.find_backwards(ci_start, in_current_line=True)
-            end = event.current_buffer.document.find(ci_end, in_current_line=True)
+            start = event.current_buffer.document.find_backwards(ci_start, in_current_line=False)
+            end = event.current_buffer.document.find(ci_end, in_current_line=False)
 
             if start is not None and end is not None:
                 offset = 0 if inner else 1
                 return CursorRegion(start + 1 - offset, end + offset)
+            else:
+                # Nothing found.
+                return CursorRegion(0)
 
     for inner in (False, True):
         for ci_start, ci_end in [('"', '"'), ("'", "'"), ("`", "`"),
                                  ('[', ']'), ('<', '>'), ('{', '}'), ('(', ')')]:
             create_ci_ca_handles(ci_start, ci_end, inner)
 
-    @change_delete_move_yank_handler('{')  # TODO: implement 'arg'
+    @change_delete_move_yank_handler('{')
     def _(event):
         """
         Move to previous blank-line separated section.
         Implements '{', 'c{', 'd{', 'y{'
         """
+        def match_func(text):
+            return not text or text.isspace()
+
         line_index = event.current_buffer.document.find_previous_matching_line(
-            lambda text: not text or text.isspace())
+            match_func=match_func, count=event.arg)
 
         if line_index:
             index = event.current_buffer.document.get_cursor_up_position(count=-line_index)
@@ -784,14 +794,17 @@ def load_vi_bindings(registry, vi_state, filter=None):
             index = 0
         return CursorRegion(index)
 
-    @change_delete_move_yank_handler('}')  # TODO: implement 'arg'
+    @change_delete_move_yank_handler('}')
     def _(event):
         """
         Move to next blank-line separated section.
         Implements '}', 'c}', 'd}', 'y}'
         """
+        def match_func(text):
+            return not text or text.isspace()
+
         line_index = event.current_buffer.document.find_next_matching_line(
-            lambda text: not text or text.isspace())
+            match_func=match_func, count=event.arg)
 
         if line_index:
             index = event.current_buffer.document.get_cursor_down_position(count=line_index)
@@ -886,16 +899,118 @@ def load_vi_bindings(registry, vi_state, filter=None):
 
     @change_delete_move_yank_handler('H')
     def _(event):
-        """ Implements 'cH', 'dH', 'H'. """
-        # Vi moves to the start of the visible region.
-        # cursor position 0 is okay for us.
-        return CursorRegion(-len(event.current_buffer.document.text_before_cursor))
+        """
+        Moves to the start of the visible region. (Below the scroll offset.)
+        Implements 'cH', 'dH', 'H'.
+        """
+        w = find_window_for_buffer_name(event.cli.layout, event.cli.current_buffer_name)
+        b = event.current_buffer
+
+        if w:
+            # When we find a Window that has BufferControl showing this window,
+            # move to the start of the visible area.
+            pos = (b.document.translate_row_col_to_index(
+                       w.render_info.first_visible_line(after_scroll_offset=True), 0) -
+                   b.cursor_position)
+
+        else:
+            # Otherwise, move to the start of the input.
+            pos = -len(b.document.text_before_cursor)
+        return CursorRegion(pos)
+
+    @change_delete_move_yank_handler('M')
+    def _(event):
+        """
+        Moves cursor to the vertical center of the visible region.
+        Implements 'cM', 'dM', 'M'.
+        """
+        w = find_window_for_buffer_name(event.cli.layout, event.cli.current_buffer_name)
+        b = event.current_buffer
+
+        if w:
+            # When we find a Window that has BufferControl showing this window,
+            # move to the center of the visible area.
+            pos = (b.document.translate_row_col_to_index(
+                       w.render_info.center_visible_line(), 0) -
+                   b.cursor_position)
+
+        else:
+            # Otherwise, move to the start of the input.
+            pos = -len(b.document.text_before_cursor)
+        return CursorRegion(pos)
 
     @change_delete_move_yank_handler('L')
     def _(event):
-        # Vi moves to the end of the visible region.
-        # cursor position 0 is okay for us.
-        return CursorRegion(len(event.current_buffer.document.text_after_cursor))
+        """
+        Moves to the end of the visible region. (Above the scroll offset.)
+        """
+        w = find_window_for_buffer_name(event.cli.layout, event.cli.current_buffer_name)
+        b = event.current_buffer
+
+        if w:
+            # When we find a Window that has BufferControl showing this window,
+            # move to the end of the visible area.
+            pos = (b.document.translate_row_col_to_index(
+                       w.render_info.last_visible_line(before_scroll_offset=True), 0) -
+                   b.cursor_position)
+
+        else:
+            # Otherwise, move to the end of the input.
+            pos = len(b.document.text_after_cursor)
+        return CursorRegion(pos)
+
+    @handle('z', '+', filter=navigation_mode|selection_mode)
+    @handle('z', Keys.ControlJ, filter=navigation_mode|selection_mode)
+    def _(event):
+        """
+        Scrolls the window to makes the current line the first line in the visible region.
+        """
+        w = find_window_for_buffer_name(event.cli.layout, event.cli.current_buffer_name)
+        b = event.cli.current_buffer
+
+        if w and w.render_info:
+            # Calculate the offset that we need in order to position the row
+            # containing the cursor in the center.
+            cursor_position_row = b.document.cursor_position_row
+
+            render_row = w.render_info.input_line_to_screen_line(cursor_position_row)
+            if render_row is not None:
+                w.vertical_scroll = max(0, render_row)
+
+
+    @handle('z', '-', filter=navigation_mode|selection_mode)
+    def _(event):
+        """
+        Scrolls the window to makes the current line the last line in the visible region.
+        """
+        w = find_window_for_buffer_name(event.cli.layout, event.cli.current_buffer_name)
+        b = event.cli.current_buffer
+
+        if w and w.render_info:
+            # Calculate the offset that we need in order to position the row
+            # containing the cursor in the center.
+            cursor_position_row = b.document.cursor_position_row
+
+            render_row = w.render_info.input_line_to_screen_line(cursor_position_row)
+            if render_row is not None:
+                w.vertical_scroll = max(0, (render_row - w.render_info.rendered_height))
+
+    @handle('z', 'z', filter=navigation_mode|selection_mode)
+    def _(event):
+        """
+        Center Window vertically around cursor.
+        """
+        w = find_window_for_buffer_name(event.cli.layout, event.cli.current_buffer_name)
+        b = event.cli.current_buffer
+
+        if w and w.render_info:
+            # Calculate the offset that we need in order to position the row
+            # containing the cursor in the center.
+            cursor_position_row = b.document.cursor_position_row
+
+            render_row = w.render_info.input_line_to_screen_line(cursor_position_row)
+            if render_row is not None:
+                w.vertical_scroll = max(0, int(render_row - w.render_info.rendered_height / 2))
 
     @change_delete_move_yank_handler('%')
     def _(event):
@@ -930,8 +1045,14 @@ def load_vi_bindings(registry, vi_state, filter=None):
         """
         Implements 'gg', 'cgg', 'ygg'
         """
-        # Move to the top of the input.
-        return CursorRegion(event.current_buffer.document.home_position)
+        d = event.current_buffer.document
+
+        if event._arg:
+            # Move to the given line.
+            return CursorRegion(d.translate_row_col_to_index(event.arg - 1, 0) - d.cursor_position)
+        else:
+            # Move to the top of the input.
+            return CursorRegion(d.get_start_of_document_position())
 
     @change_delete_move_yank_handler('g', '_')
     def _(event):
@@ -961,6 +1082,21 @@ def load_vi_bindings(registry, vi_state, filter=None):
             event.current_buffer.document.find_start_of_previous_word(
                 count=event.arg, WORD=True) or 0)
 
+    @change_delete_move_yank_handler('G')
+    def _(event):
+        """
+        Go to the end of the document. (If no arg has been given.)
+        """
+        return CursorRegion(len(event.current_buffer.document.text_after_cursor))
+
+    @handle('G', filter=HasArg())
+    def _(event):
+        """
+        If an argument is given, move to this line in the  history. (for
+        example, 15G)
+        """
+        event.current_buffer.go_to_history(event.arg - 1)
+
     @handle(Keys.Any, filter=navigation_mode)
     @handle(Keys.Any, filter=selection_mode)
     def _(event):
@@ -1010,10 +1146,22 @@ def load_vi_bindings(registry, vi_state, filter=None):
         pass
 
 
-def load_vi_system_bindings(registry, vi_state, filter=None, system_buffer_name='system'):
+def load_vi_open_in_editor_bindings(registry, vi_state, filter=None):
+    """
+    Pressing 'v' in navigation mode will open the buffer in an external editor.
+    """
+    navigation_mode = ViStateFilter(vi_state, InputMode.NAVIGATION) & ~ filters.HasSelection()
+    handle = create_handle_decorator(registry, filter)
+
+    @handle('v', filter=navigation_mode)
+    def _(event):
+        event.current_buffer.open_in_editor()
+
+
+def load_vi_system_bindings(registry, vi_state, filter=None):
     assert isinstance(vi_state, ViState)
 
-    has_focus = filters.HasFocus(system_buffer_name)
+    has_focus = filters.HasFocus(SYSTEM_BUFFER)
     navigation_mode = ViStateFilter(vi_state, InputMode.NAVIGATION) & ~ filters.HasSelection()
 
     handle = create_handle_decorator(registry, filter)
@@ -1023,7 +1171,7 @@ def load_vi_system_bindings(registry, vi_state, filter=None, system_buffer_name=
         """
         '!' opens the system prompt.
         """
-        event.cli.focus_stack.push(system_buffer_name)
+        event.cli.focus_stack.push(SYSTEM_BUFFER)
         vi_state.input_mode = InputMode.INSERT
 
     @handle(Keys.Escape, filter=has_focus)
@@ -1033,7 +1181,7 @@ def load_vi_system_bindings(registry, vi_state, filter=None, system_buffer_name=
         Cancel system prompt.
         """
         vi_state.input_mode = InputMode.NAVIGATION
-        event.cli.buffers[system_buffer_name].reset()
+        event.cli.buffers[SYSTEM_BUFFER].reset()
         event.cli.focus_stack.pop()
 
     @handle(Keys.ControlJ, filter=has_focus)
@@ -1043,7 +1191,7 @@ def load_vi_system_bindings(registry, vi_state, filter=None, system_buffer_name=
         """
         vi_state.input_mode = InputMode.NAVIGATION
 
-        system_buffer = event.cli.buffers[system_buffer_name]
+        system_buffer = event.cli.buffers[SYSTEM_BUFFER]
         event.cli.run_system_command(system_buffer.text)
         system_buffer.reset(append_to_history=True)
 
@@ -1051,7 +1199,7 @@ def load_vi_system_bindings(registry, vi_state, filter=None, system_buffer_name=
         event.cli.focus_stack.pop()
 
 
-def load_vi_search_bindings(registry, vi_state, filter=None, search_buffer_name='search'):
+def load_vi_search_bindings(registry, vi_state, filter=None, search_buffer_name=SEARCH_BUFFER):
     assert isinstance(vi_state, ViState)
 
     has_focus = filters.HasFocus(search_buffer_name)
@@ -1059,61 +1207,61 @@ def load_vi_search_bindings(registry, vi_state, filter=None, search_buffer_name=
     handle = create_handle_decorator(registry, filter)
 
     @handle('/', filter=navigation_mode)
-    @handle(Keys.ControlS)
+    @handle(Keys.ControlS, filter=~has_focus)
     def _(event):
         """
         Vi-style forward search.
         """
-        vi_state.search_direction = direction = IncrementalSearchDirection.FORWARD
+        # Set the ViState.
+        event.cli.search_state.direction = IncrementalSearchDirection.FORWARD
         vi_state.input_mode = InputMode.INSERT
 
-        if event.cli.focus_stack.current != search_buffer_name:
-            event.cli.focus_stack.push(search_buffer_name)
-
-        event.cli.buffers[event.cli.focus_stack.previous].incremental_search(direction)
+        # Focus search buffer.
+        event.cli.focus_stack.push(search_buffer_name)
 
     @handle('?', filter=navigation_mode)
-    @handle(Keys.ControlR)
+    @handle(Keys.ControlR, filter=~has_focus)
     def _(event):
         """
         Vi-style backward search.
         """
-        vi_state.search_direction = direction = IncrementalSearchDirection.BACKWARD
+        # Set the ViState.
+        event.cli.search_state.direction = IncrementalSearchDirection.BACKWARD
+
+        # Focus search buffer.
+        event.cli.focus_stack.push(search_buffer_name)
         vi_state.input_mode = InputMode.INSERT
 
-        if event.cli.focus_stack.current != search_buffer_name:
-            event.cli.focus_stack.push(search_buffer_name)
-
-        event.cli.buffers[event.cli.focus_stack.previous].incremental_search(direction)
-
     @handle(Keys.ControlJ, filter=has_focus)
     def _(event):
         """
-        Enter at the / or ? prompt.
+        Apply the search. (At the / or ? prompt.)
         """
+        input_buffer = event.cli.buffers[event.cli.focus_stack.previous]
         search_buffer = event.cli.buffers[search_buffer_name]
-        vi_state.input_mode = InputMode.NAVIGATION
+
+        # Update search state.
+        if search_buffer.text:
+            event.cli.search_state.text = search_buffer.text
+
+        # Apply search.
+        input_buffer.apply_search(event.cli.search_state)
 
         # Add query to history of search line.
-        search_buffer.add_to_history()
+        search_buffer.append_to_history()
         search_buffer.reset()
 
         # Focus previous document again.
+        vi_state.input_mode = InputMode.NAVIGATION
         event.cli.focus_stack.pop()
 
-    @handle(Keys.Any, filter=has_focus)
-    def _(event):
-        """
-        Insert text after the / or ? prompt.
-        """
-        event.cli.current_buffer.insert_text(event.data)
-
-        # Set current search search text as line search text.
-        buffer = event.cli.buffers[event.cli.focus_stack.previous]
-        buffer.set_search_text(event.cli.current_buffer.text)
+    def search_buffer_is_empty(cli):
+        """ Returns True when the search buffer is empty. """
+        return cli.buffers[search_buffer_name].text == ''
 
     @handle(Keys.Escape, filter=has_focus)
     @handle(Keys.ControlC, filter=has_focus)
+    @handle(Keys.Backspace, filter=has_focus & Condition(search_buffer_is_empty))
     def _(event):
         """
         Cancel search.
@@ -1121,24 +1269,4 @@ def load_vi_search_bindings(registry, vi_state, filter=None, search_buffer_name=
         vi_state.input_mode = InputMode.NAVIGATION
 
         event.cli.focus_stack.pop()
-        event.current_buffer.exit_isearch(restore_original_line=True)
-
         event.cli.buffers[search_buffer_name].reset()
-
-# TODO: Maybe handle insert/backspace through Change event.
-
-#    @handle(Keys.Backspace, filter=has_focus)
-#    def _(event):
-#        """
-#        Backspace at the vi-search prompt.
-#        """
-#        search_line = event.cli.buffers['search']
-#
-#        if search_line.text:
-#            search_line.delete_before_cursor()
-#            event.current_buffer.set_search_text(search_line.text)
-#        else:
-#            # If no text after the prompt, cancel search.
-#            event.current_buffer.exit_isearch(restore_original_line=True)
-#            search_line.reset()
-#            event.input_processor.pop_input_mode()
